import { createLocalVue, shallowMount, enableAutoDestroy } from '@vue/test-utils';
import flushPromises from 'flush-promises';
import EntityCollection from 'src/core/data/entity-collection.data';

import 'src/module/sw-order/component/sw-order-customer-address-select';

import 'src/app/component/form/select/base/sw-single-select';
import 'src/app/component/form/field-base/sw-base-field';
import 'src/app/component/form/field-base/sw-block-field';
import 'src/app/component/form/select/base/sw-select-base';
import 'src/app/component/form/select/base/sw-select-result-list';

import Vuex from 'vuex';

const addresses = [
    {
        id: '1',
        city: 'San Francisco',
        zipcode: '10332',
        street: 'Summerfield 27',
        country: {
            translated: {
                name: 'USA'
            }
        },
        countryState: {
            translated: {
                name: 'California'
            }
        }
    },
    {
        id: '2',
        city: 'London',
        zipcode: '48624',
        street: 'Ebbinghoff 10',
        country: {
            translated: {
                name: 'United Kingdom'
            }
        },
        countryState: {
            translated: {
                name: 'Nottingham'
            }
        }
    },
];

const customerData = {
    id: '123',
    salesChannel: {
        languageId: 'english'
    },
    billingAddressId: '1',
    shippingAddressId: '2',
    addresses: new EntityCollection(
        '/customer-address',
        'customer-address',
        null,
        null,
        [],
    ),
};

function createWrapper() {
    const localVue = createLocalVue();
    localVue.use(Vuex);

    return shallowMount(Shopware.Component.build('sw-order-customer-address-select'), {
        localVue,
        propsData: {
            customer: { ...customerData },
            value: '1',
            sameAddressLabel: 'Same address',
            sameAddressValue: '2',
        },
        stubs: {
            'sw-popover': {
                template: '<div class="sw-popover"><slot></slot></div>'
            },
            'sw-single-select': Shopware.Component.build('sw-single-select'),
            'sw-select-result-list': Shopware.Component.build('sw-select-result-list'),
            'sw-select-base': Shopware.Component.build('sw-select-base'),
            'sw-block-field': Shopware.Component.build('sw-block-field'),
            'sw-base-field': Shopware.Component.build('sw-base-field'),
            'sw-highlight-text': true,
            'sw-loader': true,
            'sw-icon': true,
            'sw-field-error': true,
            'sw-select-result': {
                props: ['item', 'index'],
                template: `<li class="sw-select-result" @click.stop="onClickResult">
                                <slot></slot>
                           </li>`,
                methods: {
                    onClickResult() {
                        this.$parent.$parent.$emit('item-select', this.item);
                    }
                }
            },
        },
        provide: {
            repositoryFactory: {
                create: () => {
                    return {
                        search: (criteria) => {
                            const collection = new EntityCollection(
                                '/customer-address',
                                'customer-address',
                                null,
                                null,
                                criteria.term !== null ? [addresses[0]] : addresses,
                            );

                            return Promise.resolve(collection);
                        },
                        get: () => Promise.resolve()
                    };
                }
            }
        }
    });
}

enableAutoDestroy(afterEach);

describe('src/module/sw-order/component/sw-order-customer-address-select', () => {
    beforeAll(() => {
    });

    it('should show address option correctly', async () => {
        const wrapper = createWrapper();
        await flushPromises();

        const billingAddressSelect = wrapper.find('.sw-select__selection');
        // Click to open result list
        await billingAddressSelect.trigger('click');

        expect(wrapper.find('li[selected="selected"]').text()).toEqual('Summerfield 27, 10332, San Francisco, California, USA');
        expect(wrapper.find('sw-highlight-text-stub').attributes().text).toEqual('Ebbinghoff 10, 48624, London, Nottingham, United Kingdom');
    });

    it('should able to set billing address same as shipping address', async () => {
        const wrapper = createWrapper();
        await flushPromises();

        const billingAddressSelect = wrapper.find('.sw-select__selection');
        // Click to open result list
        await billingAddressSelect.trigger('click');

        const sameShippingAddressOption = wrapper.find('.sw-select-result__option-same-address');
        await sameShippingAddressOption.trigger('click');

        await wrapper.vm.$nextTick();

        expect(wrapper.emitted('change')[0]).toEqual(['2']);
    });

    it('should filter entries correctly', async () => {
        const wrapper = createWrapper();
        await flushPromises();

        await wrapper.vm.searchAddress('test');

        expect(wrapper.vm.addressCriteria.term).toBe('test');
        expect(addresses[0].hidden).toBe(false);
        expect(addresses[1].hidden).toBe(true);
    });
});
